;---------------------------------------------------------------------------
;
;   INT.ASM
;
;   Include file for SB6668 Frequency counter
;
;   Contains initialisation routines and interrupt routines
;
;---------------------------------------------------------------------------

;---------------------------------------------------------------------------
;   Reset and ineterrupt vectors
;---------------------------------------------------------------------------

RESET           AJMP    INIT            Start initialising

                .NO     $0003,$FF       External interrupt 0
INT0            RETI

                .NO     $000B,$FF       Timer 0 overflow interrupt
INT_T0          AJMP    EXTEND_REF      Extend range of reference counter

                .NO     $0013,$FF       External interrupt 1
INT1            RETI

                .NO     $001B,$FF       Timer 1 overflow interrupt
INT_T1          RETI

                .NO     $0023,$FF       Serial interrupt
INT_S           AJMP    SERIAL_INT

;---------------------------------------------------------------------------
;   ROM identification data
;---------------------------------------------------------------------------

PROGRAM_ID      .AT     /SB6668 FREQUENCY COUNTER V1.0/

;---------------------------------------------------------------------------
;   ROUTINE    :  EXTEND_REF
;   FUNCTION   :  Extend reference counter range
;---------------------------------------------------------------------------

EXTEND_REF      PUSH    ACC             Save desturbed register
                MOV     A,REF_EXTEND    Increment reference counter
                ADD     A,#1             extension (BCD)
                DA      A
                MOV     REF_EXTEND,A
                POP     ACC             Restore register
                RETI

;---------------------------------------------------------------------------
;   ROUTINE    :  SERIAL_INT
;   FUNCTION   :  Serial interrupt routine
;---------------------------------------------------------------------------

SERIAL_INT      >PUSH   PSW,ACC,B,DPSEL
                JBC     RI,.RECEIVE     A byte is received
                JBC     TI,.TRANSMIT    Last byte was transmitted
                AJMP    .RETI           False interrupt (if possible)

.TRANSMIT       CLR     TI              Prevent continuous interrupts!
                JNB     SB_SELECT,.STOP_TRANS   Not selected anymore!
                JB      SB_GENERAL,.STOP_TRANS  Don't transmit if general
                JB      SOUT_HALTED,.STOP_TRANS XOFF active! Don't XMIT
                MOV     A,SOUT_TAIL     Is there anything left to be sent?
                XRL     A,SOUT_HEAD
                JZ      .STOP_TRANS     No!
                MOV     A,CR_DELAY      Is CR delay counter running?
                JNZ     .STOP_TRANS     Yes! Halt transmission for now
                MOV     DPSEL,#2        Select serial output pointer
                PUSH    DPL
                MOV     DPL,SOUT_TAIL   Copy tail pointer
                INC     SOUT_TAIL       Increment tail pointer
                MOVX    A,@DPTR         Get byte to be sent
                MOV     SBUF,A           and send it out
                POP     DPL
                MOV     LAST_SEND,A
                JNB     SLOW_MODE,.RTI  We're not in slow mode!
                CJNE    A,#CR,.RTI      It's not a CR!
                MOV     CR_DELAY,#CR_DELAY_TIME Start CR delay timer
.RTI            AJMP    .RETI
.STOP_TRANS     CLR     SOUT_ACTIVE     Don't output anymore
                AJMP    .RETI

.RECEIVE        MOV     B,SBUF          Get received byte
                JNB     B.7,.DATA       It's a data byte
                SETB    STOP_ACT        It's an address, stop SB-Bus now
                CLR     SB_SELECT       Deselect is default
                CLR     SB_GENERAL
                CLR     SOUT_HALTED
                MOV     LAST_SEND,#$FF  Special output situation
                CLR     A
                MOV     CR_COUNT,A      Clear command line counter
                MOV     CR_DELAY,A
                MOV     INLENGTH,A
                MOV     SIN_HEAD,A      Clear buffers
                MOV     SIN_TAIL,A
                MOV     SOUT_HEAD,A
                MOV     SOUT_TAIL,A
                JNB     HOLD_ACTIVE,.HOLD Hold mode not activated!
                MOV     A,HOLD_FUNCTION   Is there something in hold?
                ORL     A,HOLD_FUNCTION+1
                JNZ     .HOLD           Yes! Keep it that way
                CLR     HOLD_ACTIVE     Clear hold active flag

.HOLD           MOV     A,B
                CJNE    A,#$FF,.NOT_GENERAL It's not the general call!
                SETB    SB_SELECT           Set selected flag
                SETB    SB_GENERAL
                MOV     ERROR,#0
                AJMP    .RETI           Done!

.NOT_GENERAL    CJNE    A,MY_ADDRESS,.RETI It's not my own address
                SETB    SB_SELECT       We're selected!
                MOV     DPSEL,#2        Send => to the output buffer
                MOV     DPL,#0           (don't use system routines
                MOV     A,#'='            for theay are not reantrant)
                MOVX    @DPTR,A
                INC     DPL
                MOV     A,#'>'
                MOVX    @DPTR,A
                INC     DPL
                MOV     A,#CR
                MOVX    @DPTR,A
                MOV     SOUT_HEAD,#3
                MOV     ERROR,#0        Last error is 0

.RETI           >POP    DPSEL,B,ACC,PSW
                RETI

.DATA           JNB     SB_SELECT,.RETI
                JBC     SOUT_HALTED,.HALT  Release halted mode if set
                MOV     DPSEL,#1        Point to serial input buffer
                MOV     DPL,SIN_HEAD
                MOV     A,B
                >CMP    A,"#' '"        Is it a control character?
                JC      .CTRL           Yes!
                CALL    .PUT_CHAR       Put character in buffer
                AJMP    .RETI

.CTRL           CJNE    A,#BS,.NOTBS    Not backspace
.BSDEL          MOV     A,INLENGTH      Is something in?
                JZ      .RETI           No!
                DEC     SIN_HEAD        Decrement input buffer
                DEC     INLENGTH
                AJMP    .RETI

.NOTBS          CJNE    A,#DEL,.NOTDEL  Not DEL key!
                AJMP    .BSDEL          Handle just like BS

.NOTDEL         CJNE    A,#CR,.NOTCR    Not CR key!
                INC     CR_COUNT        Increment CR counter
                MOV     A,CR_COUNT
                JNZ     .CROK           No overflow
                DEC     CR_COUNT
.CROK           MOV     A,B
                CALL    .PUT_CHAR       Put character in buffer
                MOV     INLENGTH,#0     Clear input length counter
                AJMP    .RETI

.NOTCR          CJNE    A,#XOFF,.NOTXOFF Not XOFF key!
                SETB    SOUT_HALTED      Halt the serial output
                AJMP    .RETI

.NOTXOFF        CJNE    A,#CANCEL,.NOTCANC Not CANCEL key!
.CAN_LOOP       MOV     A,INLENGTH      Backspace until begin
                JZ      .RETI           Done!
                DEC     SIN_HEAD
                DEC     INLENGTH
                AJMP    .CAN_LOOP

.NOTCANC        CJNE    A,#ESC,.RETI    Not the escape character!
                SETB    STOP_ACT        Set stop active flag
                AJMP    .RETI

.HALT           MOV     A,B             Is it ESC?
                XRL     A,#ESC
                JNZ     .RETI           No!
                SETB    STOP_ACT        Stop all activities
                AJMP    .RETI

.PUT_CHAR       MOVX    @DPTR,A         Put new character in input buffer
                INC     DPL             Increment pointer
                MOV     A,DPL           Buffer full?
                XRL     A,SIN_TAIL
                JZ      .FULL           Yes!
                MOV     SIN_HEAD,DPL    No! Load new buffer address
                INC     INLENGTH        Increment input length
                RET
.FULL           MOV     A,B             Was it a CR?
                XRL     A,#CR
                JNZ     .RET            No!
                DEC     DPL             Put CR in buffer one position
                DEC     DPL              back in line, otherwise it will
                MOV     A,B              never be found
                MOVX    @DPTR,A
.RET            RET

;---------------------------------------------------------------------------
;   ROUTINE    :  INIT
;   FUNCTION   :  Get the system going
;---------------------------------------------------------------------------

INIT            ANL     P0,#%1111.0011  Disable decade outputs
                SETB    P3.2            Select internal reference
                MOV     SP,#STACK_BEGIN Start stack at indirect area
                MOV     AFR,#%0111.1111 Select high speed (for baud rate)
                CLR     RS0             Set main register bank to 0
                CLR     RS1

; Clear internal RAM memory
                MOV     R2,DOG_BARKED   Save watchdog reset counter
                MOV     R0,#$08         Clear memory
                MOV     R1,#120
                MOV     A,#0
.CLR_MEMORY     MOV     @R0,A
                INC     R0
                DJNZ    R1,.CLR_MEMORY

; Designate port P4 as I/O lines
                MOV     MEX1,#%0111.0111 Release address mode for P4
                MOV     MEX2,#%0111.0111
.P4_INIT_LOOP   MOV     A,MEX2          Is the extension stack full?
                JB      ACC.3,.P4_DONE  Yes! All 1's written to it!
                LCALL   .P4_INIT_LOOP   Fill extension stack with 1's
.P4_DONE        MOV     SP,#STACK_BEGIN Restore normal stack pointer

; See if reset caused by watchdog
                MOV     A,IP            Reset caused by watchdog?
                JNB     ACC.6,.NO_BARK  No!
                MOV     A,R2            Increment bark counter
                INC     A
                JNZ     .NO_OVL         No overflow!
                DEC     A
.NO_OVL         MOV     DOG_BARKED,A
.NO_BARK

; Setup buffer space
                MOV     DPSEL,#1        Setup serial data pointers
                MOV     DPH,/SIN_BUFFER
                MOV     DPSEL,#2
                MOV     DPH,/SOUT_BUFFER
                MOV     DPSEL,#0        Select main pointer again

; Init timers and interrupt system
                MOV     PCON,#%1000.0000      Double the baudrate
                MOV     TH1,#BAUD_RELOAD      Timer 1 is for baudrate
                MOV     TL1,#BAUD_RELOAD

                MOV     TMOD,#%0010.0101      T1 in mode 2, T0 in mode 1
                MOV     TCON,#%0100.0000       T0 stopped
                MOV     SCON,#%0101.0000      Start serial interface

                MOV     IP,#%0000.0010  Timer 0 has the highest priority
                MOV     IE,#%1001.0000  Enable serial interrupt

; Init SB-Bus repeat command
                MOV     REPEAT_CMD,/REP_ERROR Nothing to repeat error
                MOV     REPEAT_CMD+1,#REP_ERROR

; Start watchdog timer
                MOV     WDTREL,#64+%1000.0000 Set timeout at aprox. 1s
                SETB    WDT
                SETB    SWDT            Start watchdog

; Read initial values from NVM (initialise NVM if stamp fails)
                CALL    READ_INIT       Read initial values from NVM
                JNC     .MEM_OK         Memory is OK!
                CALL    READ_INIT       Try it again to be shure
                JNC     .MEM_OK         Memory is OK!
                CALL    INIT_NVM        Initialise NVM
.MEM_OK
                MOV     R0,#7*3         Clear the hold values
                MOV     DPTR,#HOLD_FREQ
                CLR     A
.CLR_HOLD       MOVX    @DPTR,A
                INC     DPTR
                DJNZ    R0,.CLR_HOLD    Do all bytes

; Introduce myself on the display
                CLR     A               Set index for SB-6668 or SB-6667
                MOV     C,PM6667
                MOV     ACC.0,C
                MOV     DPTR,#.HELLO_TEXT
.PM6668         CALL    SHOW_TEXT       Put text on display
                MOV     TIME_DELAY,#2000  Show name during 2 seconds
                MOV     TIME_DELAY+1,/2000

                CALL    SORT_VALUES     Sort stored tables in memory
                JMP     MAIN            Do your job

.HELLO_TEXT     .DA     #CHR_S,#CHR_B,#SG,#CHR_6,#CHR_6,#CHR_6,#CHR_8
                .DA     #CHR_S,#CHR_B,#SG,#CHR_6,#CHR_6,#CHR_6,#CHR_7

;---------------------------------------------------------------------------
;   ROUTINE    :  MAIN
;   FUNCTION   :  Main program loop
;---------------------------------------------------------------------------

MAIN            CLR     STOP_ACT        Clear stop activities flag
                CALL    POLL_SBBUS      See if a command is waiting
                CALL    DO_EVENTS       Do all other events
                JMP     MAIN

;---------------------------------------------------------------------------
;   ROUTINE    :  DO_EVENTS
;   FUNCTION   :  Do all events but SBBUS
;---------------------------------------------------------------------------

DO_EVENTS       MOV     DPSEL,#0        Default DPTR
                CALL    POLL_SOUT       See if there's anything to transmit
                CALL    SNOOZE_DOG      Keep watchdog asleep
                CALL    SYSTEM_DELAY    Handle all kinds of delays
                CALL    POLL_MENU       See if menu needs attention
                CALL    COUNTER         Service counter routines
                RET

