;---------------------------------------------------------------------------
;
;   SYSTEM.ASM
;
;   Include file for SB6668 Frequency counter
;
;   Contains general purpose system routines
;
;---------------------------------------------------------------------------

;---------------------------------------------------------------------------
;   ROUTINE    :  IIC_START
;   FUNCTION   :  Start the IIC-Bus
;---------------------------------------------------------------------------

IIC_START       SETB    P4.0            Make dataline high
                CALL    IIC_DELAY
                SETB    P4.1            Make clockline high
                CALL    IIC_DELAY
                CLR     P4.0            Make dataline low while clock high
                CALL    IIC_DELAY
                CLR     P4.1            Now make clock line low
                RET

;---------------------------------------------------------------------------
;   ROUTINE    :  IIC_STOP
;   FUNCTION   :  Stop the IIC-Bus
;---------------------------------------------------------------------------

IIC_STOP        SETB    P4.1            First rise clock line
                CALL    IIC_DELAY
                SETB    P4.0             and then the data line
                RET

;---------------------------------------------------------------------------
;   ROUTINE    :  IIC_READ
;   FUNCTION   :  Read a byte from IIC-Bus
;   INPUT      :  F0=0 if acknowledge should be given
;   OUTPUT     :  A=Read byte
;---------------------------------------------------------------------------

IIC_READ        MOV     A,#%0000.0001   Set end mark
                CLR     P4.1            Make clockline low
                CALL    IIC_DELAY
                SETB    P4.0             and dataline high
.LOOP           CALL    IIC_DELAY
                SETB    P4.1            Make clockline high
                CALL    IIC_DELAY
                MOV     C,P4.0          Get fresh databit
                RLC     A                and shift it into result
                CLR     P4.1            Make clockline low
                JNC     .LOOP           Do all 8 bits!
                CALL    IIC_DELAY
                MOV     C,F0            Copy wanted acknowledge state
                MOV     P4.0,C           to dataline
                CALL    IIC_DELAY
                SETB    P4.1            Make clockline temporary high
                CALL    IIC_DELAY
                CLR     P4.1
                CALL    IIC_DELAY       Finally make dataline low
                CLR     P4.0
                RET

;---------------------------------------------------------------------------
;   ROUTINE    :  IIC_WRITE
;   FUNCTION   :  Write a byte to the IIC-Bus
;   INPUT      :  A=databyte
;   OUTPUT     :  F0=acknowledge status (0=acknowledge)
;   DESTROYED  :  R7
;---------------------------------------------------------------------------

IIC_WRITE       MOV     R7,#8           Do 8 bits
.LOOP           CLR     P4.1            Make clockline low
                CALL    IIC_DELAY
                RLC     A               Get next bit to be transmitted
                MOV     P4.0,C           and send it out
                CALL    IIC_DELAY
                SETB    P4.1            Make clockline high
                CALL    IIC_DELAY
                DJNZ    R7,.LOOP        Do 8 bits!
                CLR     P4.1            Make clockline low for acknowledge
                CALL    IIC_DELAY
                SETB    P4.0            Make dataline high during ack
                CALL    IIC_DELAY
                SETB    P4.1            Give acknowledge pulse
                CALL    IIC_DELAY
                MOV     C,P4.0          Get acknowledge bit
                MOV     F0,C
                CLR     P4.1            Finally make clock line low
                CALL    IIC_DELAY
                CLR     P4.0            Finally make data line low
                RET

;---------------------------------------------------------------------------
;   ROUTINE    :  IIC_DELAY
;   FUNCTION   :  Don't exceed 100kHz on the IIC-Bus
;---------------------------------------------------------------------------

IIC_DELAY       NOP
                NOP
                NOP
                NOP
                RET

;---------------------------------------------------------------------------
;   ROUTINE    :  READ_NVM
;   FUNCTION   :  Read from NVM memory
;                 Uses NVM buffer to pass the data
;   UPON INPUT :  NVM_BUFFER   = Number of bytes to read (max 6)
;                 NVM_BUFFER+1 = 1st address
;   UPON OUTPUT:  NVM_BUFFER   = unchanged
;                 NVM_BUFFER+1 = Last read address + 1
;                 NVM_BUFFER+2 = 1st read data byte
;   DESTROYED  :  R0, R2, R7
;---------------------------------------------------------------------------

READ_NVM        MOV     A,NVM_BUFFER    Get number of bytes to read
                MOV     R2,A
                MOV     R0,#NVM_BUFFER+2  Read buffer
                CALL    IIC_START       Start the IC-bus
                MOV     A,#NVM_WRITE_ADDR  Send write address
                CALL    IIC_WRITE
                MOV     A,NVM_BUFFER+1  Send subaddress in NVM
                CALL    IIC_WRITE
                CALL    IIC_START       Restart the bus for reading
                MOV     A,#NVM_READ_ADDR
                CALL    IIC_WRITE

.LOOP           CLR     F0              Give acknowledge per default
                MOV     A,R2            Last byte to be read?
                DEC     A
                JNZ     .NOT_LAST       No!
                SETB    F0              Don't give acknowledge then
.NOT_LAST       CALL    IIC_READ        Read the byte
                MOV     @R0,A
                INC     R0
                INC     NVM_BUFFER+1    Increment read address
                DJNZ    R2,.LOOP        Repeat for all bytes
                CALL    IIC_STOP        Stop the IC bus
                RET

;---------------------------------------------------------------------------
;   ROUTINE    :  WRITE_NVM
;   FUNCTION   :  Write to NVM memory
;                 Uses NVM buffer to pass the data
;   UPON INPUT :  NVM_BUFFER   = Number of bytes to write (max 6)
;                 NVM_BUFFER+1 = 1st address
;                 NVM_BUFFER+2 = 1st write data byte
;   UPON OUTPUT:  NVM_BUFFER   = unchanged
;                 NVM_BUFFER+1 = Last written address +1
;                 C=1 if error(s) occurred during write process
;   DESTROYED  :  R0,R1,R2,R3,R4,R5,R7
;---------------------------------------------------------------------------

WRITE_NVM       MOV     R3,#0           Clear error flag
                MOV     A,NVM_BUFFER    Get number of bytes to be written
                MOV     R2,A
                MOV     R0,#NVM_BUFFER+2 Point to first data byte

.LOOP           MOV     R4,#3           Retry writing 3 times!
.RETRY_LOOP     CALL    IIC_START       Start IC bus
                MOV     A,#NVM_WRITE_ADDR  Address the NVM memory
                CALL    IIC_WRITE
                MOV     A,NVM_BUFFER+1  Send sub address in NVM
                CALL    IIC_WRITE
                MOV     A,@R0           Send byte to be written
                CALL    .WRITE_BYTE     Start writing and wait for ack

                CALL    IIC_START       Read the databyte back
                MOV     A,#NVM_WRITE_ADDR
                CALL    IIC_WRITE
                MOV     A,NVM_BUFFER+1
                CALL    IIC_WRITE
                CALL    IIC_START
                MOV     A,#NVM_READ_ADDR
                CALL    IIC_WRITE
                SETB    F0              Don't send an acknowledge
                CALL    IIC_READ
                CALL    IIC_STOP
                XRL     A,@R0           Same as byte to be written?
                JZ      .EQUAL          Yes!
                DJNZ    R4,.RETRY_LOOP  Try again!
                MOV     A,R3            Set error flag
                ORL     A,#%1000.0000
                MOV     R3,A

.EQUAL          INC     R0              Point to next byte
                INC     NVM_BUFFER+1    Increment NVM address
                DJNZ    R2,.LOOP        Write all bytes this way
                MOV     A,R3            Get error bit
                MOV     C,ACC.7         C=1 if error occurred
                RET

.WRITE_BYTE     CALL    IIC_WRITE       Write the databyte
                CALL    IIC_STOP        Start the burning process
                MOV     R1,#100         Setup maximum delay time

.WRITE_LOOP     MOV     R5,#0           Waste some time
.WASTE          DJNZ    R5,.WASTE
                CALL    IIC_START       See NVM ready (if so it will send
                MOV     A,#NVM_WRITE_ADDR  an acknowledge back)
                CALL    IIC_WRITE
                JNB     F0,.DONE        Acknowledge received! He's done
                DJNZ    R1,.WRITE_LOOP  Keep waiting
.DONE           CALL    IIC_STOP        Stop the IC bus again
                RET

;---------------------------------------------------------------------------
;   ROUTINE    :  READ_INIT
;   FUNCTION   :  Read initial system status from NVM memory
;---------------------------------------------------------------------------

READ_INIT       MOV     NVM_BUFFER,#6   Read 6 bytes (id, address, options)
                MOV     NVM_BUFFER+1,#NVM_ID
                CALL    READ_NVM        Read the bytes
                MOV     R0,#NVM_BUFFER+2 See if ID matches
                MOV     DPTR,#NVM_DFLT_TABLE
                MOV     R1,#4           4 characters must match
.ID_CHECK       CLR     A
                MOVC    A,@A+DPTR
                XRL     A,@R0           Do they match?
                JNZ     .ID_FAIL        No! Memory lost!
                INC     R0              Increment pointers
                INC     DPTR
                DJNZ    R1,.ID_CHECK    Do 4 characters
                MOV     A,NVM_BUFFER+6  Setup SB-Bus address
                ORL     A,#%1000.0000
                MOV     MY_ADDRESS,A
                MOV     A,NVM_BUFFER+7  Setup options
                MOV     C,ACC.0         Setup PM6667 or PM6668 flag
                MOV     PM6667,C
                MOV     C,ACC.1         Set ext reference available flag
                MOV     OPTION_EXT_REF,C

                MOV     DPTR,#VALUE_TABLE Read the values from NVM
                MOV     NVM_BUFFER,#5   Read 5 bytes at a time
                MOV     R3,#50          Read 50 values
.READ_VALUES    CALL    READ_NVM        Read values
                MOV     R0,#NVM_BUFFER+2 Point to read values
                MOV     R1,#5           Copy 5 bytes
.COPY_VALUE     MOV     A,@R0            Copy 1 byte
                MOVX    @DPTR,A
                INC     R0              Increment pointers
                INC     DPTR
                DJNZ    R1,.COPY_VALUE  Do all 5 bytes
                DJNZ    R3,.READ_VALUES Do all 50 values
                CLR     C               Indicate memory OK
                RET

.ID_FAIL        SETB    C               Indicate memory lost
                RET

;---------------------------------------------------------------------------
;   ROUTINE    :  INIT_NVM
;   FUNCTION   :  Initialise NVM memory, for memory was lost
;---------------------------------------------------------------------------

INIT_NVM        MOV     DPTR,#.OOPS_TEXT Show OOPS in the display to signal
                CLR     A                 memory lost situation
                CALL    SHOW_TEXT
                MOV     NVM_BUFFER,#1   Write 1 byte at a time
                MOV     R0,#0           Copy 256 bytes from default table
                MOV     R1,#0            to NVM memory
                MOV     DPTR,#NVM_DFLT_TABLE
.WRITE_NVM      MOV     NVM_BUFFER+1,R0 Setup write address
                MOV     A,R0            Get byte to be written
                MOVC    A,@A+DPTR
                MOV     NVM_BUFFER+2,A
                >PUSH   B0_R0,B0_R1
                CALL    WRITE_NVM       Write byte to NVM memory
                JNC     .NO_WRITE_ERROR No write error occurred!
                SETB    IIC_ERROR       Set IC error bit
.NO_WRITE_ERROR >POP    B0_R1,B0_R0
                INC     R0              Increment address pointers
                CALL    SNOOZE_DOG      Keep the dog asleep
                DJNZ    R1,.WRITE_NVM   Do all 256
                JMP     READ_INIT

.OOPS_TEXT      .DA     #BLANK,#CHR_0,#CHR_0,#CHR_P,#CHR_S,#BLANK,#BLANK

;---------------------------------------------------------------------------
;   ROUTINE    :  NVM_DFLT_TABLE
;   FUNCTION   :  NVM memory default contents (after memory lost)
;---------------------------------------------------------------------------

NVM_DFLT_TABLE  .AS     /6668/          1st four bytes of NVM
                .DA     #254            Default SB-Bus address
                .DA     #%0000.0000     Default option byte

                .HS     .00.00.00.02.00         Scale 2
                .HS     .00.00.00.03.00         Scale 3
                .HS     .00.00.00.04.00         Scale 4
                .HS     .00.00.00.08.00         Scale 8
                .HS     .00.00.00.01.01         Scale 10
                .HS     .00.00.50.01.01         Scale 15
                .HS     .00.00.60.01.01         Scale 16
                .HS     .00.00.20.03.01         Scale 32
                .HS     .00.00.40.06.01         Scale 64
                .HS     .00.00.28.01.02         Scale 128
                .HS     .00.00.56.02.02         Scale 256
                .HS     .00.00.12.05.02         Scale 512

                .HS     .00.00.45.04.85         Offset 445kHz
                .HS     .00.00.50.04.85         Offset 450kHz
                .HS     .00.00.55.04.85         Offset 455kHz
                .HS     .00.00.50.05.86         Offset 5.5MHz
                .HS     .00.00.74.05.86         Offset 5.74MHz
                .HS     .00.00.00.06.86         Offset 6MHz
                .HS     .00.00.50.06.86         Offset 6.5MHz
                .HS     .00.00.07.01.87         Offset 10.07MHz
                .HS     .00.00.89.03.87         Offset 38.9MHz

                .NO     NVM_DFLT_TABLE+256,$FF Fill rest of EEPROM with $FF

;---------------------------------------------------------------------------
;   ROUTINE    :  SNOOZE_DOG
;   FUNCTION   :  Keep watchdog asleep
;---------------------------------------------------------------------------

SNOOZE_DOG      CLR     EA              Clear global interrupts
                SETB    WDT             Initiate refresh
                SETB    SWDT            Restart watchdog timer
                SETB    EA              Enable all interrupts again
                RET

;---------------------------------------------------------------------------
;   ROUTINE    :  SEND_DEC
;   FUNCTION   :  Send number in A as decimal number to SB-Bus
;---------------------------------------------------------------------------

SEND_DEC        CLR     F1              Clear leading 0 flag
                MOV     B,#100          Divide by 100 first
                DIV     AB
                JZ      .LEAD1          Don't print leading 0!
                SETB    F1              Set leading 0 flag
                ORL     A,#'0'          Make it ASCII
                CALL    COUT
.LEAD1          MOV     A,B             Divide by 10 now
                MOV     B,#10
                DIV     AB
                JNZ     .LEAD2          Not a leading 0!
                JNB     F1,.LEAD3       Still leading 0!
.LEAD2          ORL     A,#'0'          Make it ASCII
                CALL    COUT
.LEAD3          MOV     A,B             Finally send last digit
                ORL     A,#'0'
                CALL    COUT
                RET

;---------------------------------------------------------------------------
;   ROUTINE    :  SEND_HEX
;   FUNCTION   :  Send number in A as hex number to SB-Bus
;---------------------------------------------------------------------------

SEND_HEX        MOV     B,A             Save lower nibble
                SWAP    A               Send upper nibble 1st
                CALL    .CONVERT
                MOV     A,B             And then lower nibble
.CONVERT        ANL     A,#%0000.1111
                ORL     A,#'0'          Make it an ASCII sign
                >CMP    A,#'9'+1
                JC      .OK             Digit is 0...9!
                ADD     A,#7            Make it A...F
.OK             CALL    COUT
                RET

;---------------------------------------------------------------------------
;   ROUTINE    :  UPDATE_DSP
;   FUNCTION   :  Update display       Left digit 1st
;---------------------------------------------------------------------------

UPDATE_DSP      ANL     P0,#%1111.0011    Disable decimal counter outputs
                MOV     R0,#DSP_BUFFER+7  Point to annunciator byte
                MOV     C,REMOTE_MODE     Switch Lowbat indicator on if
                MOV     A,@R0              remote mode is active
                MOV     ACC.3,C
                MOV     @R0,A
                MOV     R0,#DSP_BUFFER    Start with leftmost display
                ANL     P1,#%1111.0000    Clear all 3 dsp enables and data
                ANL     P0,#%1110.1111    Make clock line low
                SETB    P1.2              Enable display controller EN1
                CALL    .DISP1          Send low startbit
                CALL    .DISP8          Send 2 full digits to the display
                CALL    .DISP8           controller
                MOV     A,@R0           Followed by 4 last bits
                INC     R0
                CALL    .DISP4          Followed by 4 last bits
                MOV     R3,A
                CLR     P1.2            Release enable EN1
                SETB    P1.1            Enable controller EN2
                CALL    .DISP1          Send low startbit
                MOV     A,R3
                CALL    .DISP4          Send last 4 bits too
                CALL    .DISP8          Send 2 full digits to the display
                CALL    .DISP8           controller
                CLR     P1.1            Release enable EN2
                SETB    P1.0            Enable controller EN3
                CALL    .DISP1          Send low startbit
                CALL    .DISP8          Send 2 full digits to the display
                CALL    .DISP8           controller
                MOV     A,@R0           Send 4 bits of annunciator data
                CALL    .DISP4
                CLR     P1.0            Release enable EN3

.DISP1          CLR     A               Send final load pulse to last
                CPL     A
                MOV     R1,#1            controller

.SEND_DISP      MOV     C,ACC.0         Copy inverted bit to the dataline
                CPL     C
                MOV     P1.3,C
                RR      A               Prepare next bit to be send
                SETB    P0.4            Pulse the clock line for 5s
                MOV     R2,#6
.DELAY          DJNZ    R2,.DELAY
                CLR     P0.4            Make clock line low again
                NOP
                NOP
                NOP
                NOP
                DJNZ    R1,.SEND_DISP   Do all wanted bits!
                RET

.DISP4          MOV     R1,#4           Send 4 bits
                JMP     .SEND_DISP

.DISP8          MOV     A,@R0           Send next digit completely
                INC     R0
                MOV     R1,#8
                JMP     .SEND_DISP

;---------------------------------------------------------------------------
;   ROUTINE    :  BCD_2_7SEGM
;   FUNCTION   :  Convert BCD digit to seven segment digit
;---------------------------------------------------------------------------

BCD_2_7SEGM     ANL     A,#%0000.1111   Mask off irrelevant bits
                MOV     DPTR,#.TABLE    Load table start address
                MOVC    A,@A+DPTR
                RET

.TABLE          .DA     #CHR_0
                .DA     #CHR_1
                .DA     #CHR_2
                .DA     #CHR_3
                .DA     #CHR_4
                .DA     #CHR_5
                .DA     #CHR_6
                .DA     #CHR_7
                .DA     #CHR_8
                .DA     #CHR_9
                .DA     #CHR_A
                .DA     #CHR_B
                .DA     #CHR_C
                .DA     #CHR_D
                .DA     #CHR_E
                .DA     #CHR_F

;---------------------------------------------------------------------------
;   ROUTINE    :  SYSTEM_DELAY
;   FUNCTION   :  Update all kinds of delay counters
;---------------------------------------------------------------------------

SYSTEM_DELAY    PUSH    ACC             Save counter register
                MOV     C,P3.3          Complement test pin
                CPL     C
                MOV     P3.3,C
                MOV     A,#TUNE_TIME    Setup delay counter for 1ms
.DELAY          NOP
                NOP
                NOP
                NOP
                NOP
                NOP
                NOP
                DJNZ    ACC,.DELAY      Time not over yet!

; Handle CR delay time
                MOV     A,CR_DELAY      Is delay timer running?
                JZ      .NOT_CR_DELAY   No! Don't bother then
                DEC     CR_DELAY        Decrement delay timer
.NOT_CR_DELAY

; Handle counter delay
                MOV     A,TIME_DELAY    Is timer delay running?
                JNZ     .DEC_COUNT_LOW  Yes! Decrement only low part
                MOV     A,TIME_DELAY+1
                JZ      .NOT_COUNT_DEL  No! Don't bother then
                DEC     TIME_DELAY+1    Decrement both high and low
.DEC_COUNT_LOW  DEC     TIME_DELAY      Decrement low byte of counter
.NOT_COUNT_DEL

; Handle keyboard scan
                MOV     C,P3.1          Get key from input pin
                CPL     C                and complement it
                MOV     ACC.0,C         Did state change since last time
                XRL     A,KEY_FLAGS
                JNB     ACC.0,.OTHER_KEY No!
                MOV     LAST_KEY_STATE,C Copy last key status
                SJMP    .SKIP_KEY
.OTHER_KEY      MOV     KEY_STATE,C     Copy current key status
.SKIP_KEY

                POP     ACC
                RET

;---------------------------------------------------------------------------
;   ROUTINE    :  CLR_FP_ACCU
;   FUNCTION   :  CLR floating point ACCU (R0 points to FP accu)
;---------------------------------------------------------------------------

CLR_FP_ACCU     CLR     A
                PUSH    B0_R0
                MOV     R2,#6           Clear only 6 digit pairs
.LOOP           MOV     @R0,A           Clear the ACCU
                INC     R0
                DJNZ    R2,.LOOP        Do all bytes
                POP     B0_R0
                RET

;---------------------------------------------------------------------------
;   ROUTINE    :  BIN2BCD
;   FUNCTION   :  Convert binary value in A to BCD result (FP_ACCU_B)
;---------------------------------------------------------------------------

BIN2BCD         MOV     R2,#8           Convert a total of 8 bits
                MOV     R3,A            Save byte to be converted
.LOOP           MOV     R0,#FP_ACCU_B   Use FP_ACCU_B
                MOV     A,@R0           Multiply temp result by 2
                ADD     A,@R0
                DA      A
                MOV     R4,A            Save temp result
                PUSH    PSW              and the carry
                MOV     A,R3            Get bit to be converted
                RL      A
                MOV     R3,A
                MOV     C,ACC.0
                MOV     A,R4            Insert it to result
                MOV     ACC.0,C
                MOV     @R0,A           Save LSB result
                POP     PSW
                MOV     R4,#3           Add 3 more BCD pairs
.REST           INC     R0
                MOV     A,@R0           Multiply BCD pair by 2
                ADDC    A,@R0
                DA      A
                MOV     @R0,A
                DJNZ    R4,.REST        Do them all!
                DJNZ    R2,.LOOP        Do all 8 bits
                RET

;---------------------------------------------------------------------------
;   ROUTINE    :  FP_SHIFT_LEFT
;   FUNCTION   :  Shift FP ACCU (R0) 1 digit to the left
;---------------------------------------------------------------------------

FP_SHIFT_LEFT   MOV     R4,#0           Shift 0 in from the right
                MOV     R5,#6           Do 6 pairs
                PUSH    B0_R0

.LOOP           MOV     A,@R0           Get BCD pair and swap lower and
                SWAP    A                upper digit
                MOV     @R0,A
                MOV     A,R4            Swap highest digit from previous
                XCHD    A,@R0            byte with former highest digit
                MOV     R4,A             of current byte

                INC     R0
                DJNZ    R5,.LOOP        Do all pairs
                POP     B0_R0
                RET

;---------------------------------------------------------------------------
;   ROUTINE    :  FP_SHIFT_RIGHT
;   FUNCTION   :  Shift FP ACCU (R0) 1 digit to the right
;---------------------------------------------------------------------------

FP_SHIFT_RIGHT  MOV     R4,#0           Shift 0 in from the left
                MOV     R5,#6           Do 6 pairs
                PUSH    B0_R0
                MOV     A,R0            Start at the other end
                ADD     A,#5
                MOV     R0,A

.LOOP           MOV     A,R4            Swap lowest digit from this pair
                XCHD    A,@R0            with previous temp digit
                MOV     R4,A
                MOV     A,@R0           Put new pair in propperorder
                SWAP    A
                MOV     @R0,A

                DEC     R0
                DJNZ    R5,.LOOP        Do all pairs
                POP     B0_R0
                RET

;---------------------------------------------------------------------------
;   ROUTINE    :  ADD_FP_NORM
;   FUNCTION   :  Add FP (R1) to FP (R0)
;---------------------------------------------------------------------------

ADD_FP_NORM     CALL    MATCH_EXPONENTS Match both exponents
                CLR     SUB_ZERO        Set flags propperly
                SETB    SUB_SIGN
                CALL    ADD_FP          Add both Accu's
                CALL    NORMALISE       Normalise result
                RET

;---------------------------------------------------------------------------
;   ROUTINE    :  ADD_FP
;   FUNCTION   :  Add FP (R1) to FP (R0)
;                 Both exponents must be equal
;                 Result is not normalised
;---------------------------------------------------------------------------

ADD_FP          >PUSH   B0_R0,B0_R1
                MOV     R2,#6           Do 6 BCD pairs
                CLR     C               ADDC is used
.LOOP           MOV     A,@R1           Add two BCD pairs together
                ADDC    A,@R0
                DA      A
                MOV     @R0,A
                INC     R0              Point to next pairs
                INC     R1
                DJNZ    R2,.LOOP        Do them all!
                >POP    B0_R1,B0_R0
                RET

;---------------------------------------------------------------------------
;   ROUTINE    :  SUB_FP_NORM
;   FUNCTION   :  Subtract FP (R1) from FP (R0)
;---------------------------------------------------------------------------

SUB_FP_NORM     CALL    MATCH_EXPONENTS Match both exponents
                CALL    SUBTRACT_FP     Subtract Accu's
                CALL    NORMALISE       Normalise result
                RET

;---------------------------------------------------------------------------
;   ROUTINE    :  SUBTRACT_FP
;   FUNCTION   :  Subtract FP (R1) from FP (R0)
;                 Both exponents must be equal
;                 Result is not normalised
;---------------------------------------------------------------------------

SUBTRACT_FP     >PUSH   B0_R0,B0_R1
                MOV     R2,#6           Do 6 BCD pairs
                SETB    C               Set carry (no borrow state)
.LOOP           MOV     A,#$99          (-1 in BCD format)
                ADDC    A,#$00          Add the borrow
                MOV     R3,A            Save the borrow status
                MOV     A,@R1           Form 2s complement value
                CPL     A
                INC     A
                ADD     A,R3            Add borrow if needed
                ADD     A,@R0           Make the actual subtraction
                DA      A                in BCD format
                MOV     @R0,A
                INC     R0              Increment both pointers
                INC     R1
                DJNZ    R2,.LOOP        Do all 6 bytes!
                CLR     SUB_ZERO        Value is not 0 by default
                MOV     SUB_SIGN,C      Copy borrow to sign value
                MOV     R2,#6           Check 6 digit pairs 0 value
                CLR     A
.ZERO           DEC     R0              Decrement pointer
                ORL     A,@R0           Test byte
                DJNZ    R2,.ZERO        Do all 6 bytes
                JNZ     .DONE           Value is not zero!
                SETB    SUB_ZERO        Value is zero!
.DONE           >POP    B0_R1,B0_R0
                RET

;---------------------------------------------------------------------------
;   ROUTINE    :  MUL_FP_NORM
;   FUNCTION   :  Multiply FP_A to FP_C
;---------------------------------------------------------------------------

MUL_FP_NORM     MOV     A,FP_EXP_A      Add both exponents
                ADD     A,FP_EXP_C
                MOV     FP_EXP_C,A
                MOV     R0,#FP_ACCU_C   Copy accu C to B
                MOV     R1,#FP_ACCU_B
                CALL    COPY_FP_ACCU
                CALL    CLR_FP_ACCU     Clear result accu C, but not EXP_C!
                MOV     R1,#FP_ACCU_A
                MOV     R3,#11          Do a total of 11 digits
.LOOP           MOV     A,FP_ACCU_B+5   Get upper digit
                ANL     A,#%0000.1111
                JZ      .ZERO           Digit is zero! Nothing to do
                MOV     R6,A            Use it as a counter
.ADD_LOOP       MOV     R0,#FP_ACCU_C   Add Accu A to Accu C
                CALL    ADD_FP
                MOV     A,FP_ACCU_C+5   Need normalising?
                ANL     A,#%1111.0000
                JZ      .SKIP           No!
                CALL    FP_SHIFT_RIGHT  Shift FP Accu C right
                INC     FP_EXP_C         and increment it's exponent
                MOV     R0,#FP_ACCU_A   Shift FP Accu A right 1 digit
                CALL    FP_SHIFT_RIGHT   to compensate
                DEC     R3              Decrement digit counter
                MOV     A,R3
                JNZ     .SKIP           Result not 0!
                INC     R3              Prevent digit from becomming 0
.SKIP           DJNZ    R6,.ADD_LOOP
.ZERO           MOV     R0,#FP_ACCU_B   Shift FP Accu B left 1 digit
                CALL    FP_SHIFT_LEFT
                MOV     R0,#FP_ACCU_A   Shift FP Accu A right 1 digit
                CALL    FP_SHIFT_RIGHT
                DJNZ    R3,.LOOP        Do next digit
                RET

;---------------------------------------------------------------------------
;   ROUTINE    :  DIV_FP_NORM
;   FUNCTION   :  Divide FP C by FP A and put result in FP C
;---------------------------------------------------------------------------

DIV_FP_NORM     MOV     R0,#FP_ACCU_C   Copy accu C to B
                MOV     R1,#FP_ACCU_B
                CALL    COPY_FP_ACCU

;---------------------------------------------------------------------------
;   ROUTINE    :  DIVIDE_FP
;   FUNCTION   :  Divide FP_ACCU_B by FP_ACCU_A, result in FP_ACCU_C
;---------------------------------------------------------------------------

DIVIDE_FP       MOV     R0,#FP_ACCU_C   Clear result Accu
                CALL    CLR_FP_ACCU

                MOV     A,FP_EXP_B      Subtract exponents from each other
                CLR     C
                SUBB    A,FP_EXP_A
                MOV     FP_EXP_C,A

                MOV     R7,#8           Do 8 digits (1 used for rounding)
                MOV     R6,#FP_ACCU_C+5 Point to result register (MSD-1)
                CLR     F0               (F0 indicates upper part if 1)
                MOV     R0,#FP_ACCU_B   Subtract A from B to B
                MOV     R1,#FP_ACCU_A
                CALL    SUBTRACT_FP     Try to subtract
                JC      .OK             Subtraction worked
                CALL    ADD_FP          Subtraction didn't work
                DEC     FP_EXP_C        Shift B left and keep value same

.NEXT_DIGIT     CALL    FP_SHIFT_LEFT   Do next decade
.AGAIN          CALL    SUBTRACT_FP     Try to subtract
                JNC     .NOT_OK         Subtraction didn't work!
.OK             MOV     A,R0            Get result pointer
                XCH     A,R6
                MOV     R0,A
                JNB     F0,.LOW_PART    Do lower part of byte!
                MOV     A,#$10          Increment higher part of byte
                ADD     A,@R0
                MOV     @R0,A
                SJMP    .RESTORE_PNTRS  Try it again!
.LOW_PART       INC     @R0
.RESTORE_PNTRS  MOV     A,R0            Restore pointers
                XCH     A,R6
                MOV     R0,A
                JMP     .AGAIN          Try it again!

.NOT_OK         CALL    ADD_FP          Restore value, subtraction failed
                JB      F0,.CPL_F0      No need to decrement pointer!
                DEC     R6              Decrement register pointer
.CPL_F0         CPL     F0              Complement F0
                DJNZ    R7,.NEXT_DIGIT  Do next digit
                RET

;---------------------------------------------------------------------------
;   ROUTINE    :  MATCH_EXPONENTS
;   FUNCTION   :  Match exponents of 2 FP Accus, pointed to by R1 and R0
;---------------------------------------------------------------------------

MATCH_EXPONENTS MOV     A,R0            Get 1st exponent
                MOV     R2,A
                ADD     A,#6
                MOV     R0,A
                MOV     A,@R0
                XCH     A,R2
                MOV     R0,A
                MOV     A,R1            Get 2nd exponent
                MOV     R3,A
                ADD     A,#6
                MOV     R1,A
                MOV     A,@R1
                XCH     A,R3
                MOV     R1,A

.LOOP           MOV     A,R2            Compare both values
                CJNE    A,B0_R3,.NOT_EQ They're not equal yet!
                MOV     A,R0            Copy new exponent values
                MOV     R4,A             starting with exponent R0
                ADD     A,#6
                MOV     R0,A
                MOV     A,R2
                MOV     @R0,A
                MOV     A,R4
                MOV     R0,A
                MOV     A,R1            And finally exponent R1
                MOV     R4,A
                ADD     A,#6
                MOV     R1,A
                MOV     A,R3
                MOV     @R1,A
                MOV     A,R4
                MOV     R1,A
                RET                     Done!

.NOT_EQ         JNC     .SHIFT_R1       Shift Accu with smallest exponent
                CALL    FP_SHIFT_RIGHT  Shift Accu (R0) 1 digit right
                INC     R2              Increment exponent
                JMP     .LOOP
.SHIFT_R1       MOV     A,R0            Exchange pointers
                XCH     A,R1
                MOV     R0,A
                CALL    FP_SHIFT_RIGHT  Shift Accu (R1) 1 digit right
                MOV     A,R0            Exchange pointers again
                XCH     A,R1
                MOV     R0,A
                INC     R3              Increment exponent
                JMP     .LOOP

;---------------------------------------------------------------------------
;   ROUTINE    :  NORMALISE
;   FUNCTION   :  Normalise result in FP Accu, pointed to by R0
;---------------------------------------------------------------------------

NORMALISE       MOV     A,R0            Save pointer
                MOV     R7,A
                MOV     R6,#12          Do a maximum of 12 digits
                ADD     A,#5            Point at highest nibble pair
                MOV     R0,A
                MOV     A,@R0            and get it
                ANL     A,#%1111.0000   Upper nibble 0?
                JZ      .LEFT           Yes! Shift left
                INC     R0              Increment exponent value
                INC     @R0
                MOV     A,R7            Restore pointer
                MOV     R0,A
                CALL    FP_SHIFT_RIGHT  Shift result 1 position right
                RET
.LEFT           MOV     A,R7            See if upper digit <> 0
                ADD     A,#5
                MOV     R0,A
                MOV     A,@R0
                JNZ     .DONE           It is! We're done
                INC     R0              Decrement exponent
                DEC     @R0
                MOV     A,R7            Restore pointer
                MOV     R0,A
                CALL    FP_SHIFT_LEFT   Shift result 1 position left
                DJNZ    R6,.LEFT        Repeat a maximum 12 times
                MOV     A,R7            Result is 0, make exponent 0 too
                ADD     A,#6
                MOV     R0,A
                CLR     A
                MOV     @R0,A
.DONE           MOV     A,R7            Restore pointer
                MOV     R0,A
                RET

;---------------------------------------------------------------------------
;   ROUTINE    :  SHOW_TEXT
;   FUNCTION   :  Show a constant text, pointed to by DPTR, on the display
;                 A is an index, telling what string is to be printed
;---------------------------------------------------------------------------

SHOW_TEXT       MOV     B,#7            Multiply index with string length
                MUL     AB               and add it to the start of the
                ADD     A,DPL            text table
                MOV     DPL,A
                MOV     A,DPH
                ADDC    A,#0
                MOV     DPH,A
                MOV     R0,#DSP_BUFFER  Setup display buffer
                MOV     R1,#7           Do 7 characters
.LOOP           CLR     A               Get a display immage
                MOVC    A,@A+DPTR
                MOV     @R0,A            and put in the display buffer
                INC     DPTR            Increment pointers
                INC     R0
                DJNZ    R1,.LOOP        Do all characters
                MOV     @R0,#0          Clear all annunciators
                CALL    UPDATE_DSP
                RET

;---------------------------------------------------------------------------
;   ROUTINE    :  MEM_2_FP_ACCU
;   FUNCTION   :  Copy memory FP value to any FP Accu
;   UPON ENTRY :  R0=Pointer to the FP accu
;                 DPTR=Pointer to desired memory location
;   UPON EXIT  :  C=0 if scaling, C=1 if offset
;---------------------------------------------------------------------------

MEM_2_FP_ACCU   CLR     A               Clear 4 LS digits
                MOV     @R0,A
                INC     R0
                MOV     @R0,A
                INC     R0
                MOV     R7,#4           Copy 4 digit pairs
.LOOP           MOVX    A,@DPTR         Get digit pair
                MOV     @R0,A            and copy it to accu
                INC     DPTR            Increment pointers
                INC     R0
                DJNZ    R7,.LOOP        Do all 4 digit pairs!
                MOVX    A,@DPTR         Get exponent byte
                INC     DPTR
                MOV     C,ACC.7         Keep flag byte in C
                PUSH    PSW
                MOV     C,ACC.6         Get rid of flag bit
                MOV     ACC.7,C          with respect to sign extend
                POP     PSW
                MOV     @R0,A           Save exponent too
                RET

;---------------------------------------------------------------------------
;   ROUTINE    :  FP_ACCU_2_MEM
;   FUNCTION   :  Copy any FP Accu to memory FP value
;   UPON ENTRY :  R0=Pointer to the FP accu
;                 DPTR=Pointer to desired memory location
;                 C=Scale/Offset flag (0=scale, 1=offset)
;---------------------------------------------------------------------------

FP_ACCU_2_MEM   PUSH    PSW             Save scale/offset flag
                INC     R0              Skip 4 LS digits
                INC     R0
                MOV     R7,#4           Copy 4 digit pairs
.LOOP           MOV     A,@R0           Get digit pair
                MOVX    @DPTR,A          and copy it to RAM
                INC     R0              Increment pointers
                INC     DPTR
                DJNZ    R7,.LOOP        Do all 4 digit pairs
                POP     PSW             Restore scale/offset flag
                MOV     A,@R0           Copy exponent byte
                MOV     ACC.7,C         Set scale/offset flag propperly
                MOVX    @DPTR,A
                INC     DPTR
                RET

;---------------------------------------------------------------------------
;   ROUTINE    :  COPY_FP_ACCU
;   FUNCTION   :  Copy FP accu (R0) to FP accu (R1)
;---------------------------------------------------------------------------

COPY_FP_ACCU    >PUSH   B0_R0,B0_R1     Save pointers
                MOV     R2,#7           Copy the 7 byte accu
.LOOP           MOV     A,@R0
                MOV     @R1,A
                INC     R0
                INC     R1
                DJNZ    R2,.LOOP        Do all bytes!
                >POP    B0_R1,B0_R0     Restore pointer
                RET

;---------------------------------------------------------------------------
;   ROUTINE    :  DEL_VALUE
;   FUNCTION   :  Delete the value in PARAMETERS+1..+5 from memory and NVM
;---------------------------------------------------------------------------

DEL_VALUE       MOV     DPTR,#VALUE_TABLE  Point to the value table
                MOV     R2,#50          Scan 50 values
.MAIN_LOOP      MOV     R0,#PARAMETERS+1 Point to value to look for
                MOV     R3,DPL          Save current DPTR address
                MOV     R1,#5           Compare 5 bytes
.SCAN_LOOP      MOVX    A,@DPTR         Are they the same?
                XRL     A,@R0
                JNZ     .NOT_EQUAL      No!
                INC     R0              Test next bytes too
                INC     DPTR
                DJNZ    R1,.SCAN_LOOP
                MOV     DPL,R3          Found it! Now delete it
                MOV     R1,#5
                MOV     A,#$FF
.DELETE_LOOP    MOVX    @DPTR,A         Clear this byte
                INC     DPTR
                DJNZ    R1,.DELETE_LOOP
                SJMP    .DELETE_NVM     Now delete it from NVM

.NOT_EQUAL      MOV     A,R3            Try next value
                ADD     A,#5
                MOV     R3,A
                MOV     DPL,A
                DJNZ    R2,.MAIN_LOOP   Try all 50 values
                MOV     ERROR,#14       Exit with value not found error
                RET

.DELETE_NVM     MOV     NVM_BUFFER,#5   Read 5 bytes at a time and
                MOV     NVM_BUFFER+1,#NVM_ARRAY  start at the top of array
                MOV     R3,#50          Scan a total of 50 values
.NVM_SCAN_LOOP  CALL    READ_NVM        Read a value from NVM memory
                MOV     R0,#NVM_BUFFER+2 Setup pointers to compare the values
                MOV     R1,#PARAMETERS+1
                MOV     R2,#5           Compare a total of 5 bytes
.NVM_CMP_LOOP   MOV     A,@R0           Compare bytes
                XRL     A,@R1
                JNZ     .NVM_NOT_EQUAL  They don't match
                INC     R0
                INC     R1
                DJNZ    R2,.NVM_CMP_LOOP Compare all bytes
                MOV     A,NVM_BUFFER+1  Point to found value again
                CLR     C
                SUBB    A,#5
                MOV     NVM_BUFFER+1,A
                MOV     A,#$FF          Erase NVM buffer
                MOV     NVM_BUFFER+2,A
                MOV     NVM_BUFFER+3,A
                MOV     NVM_BUFFER+4,A
                MOV     NVM_BUFFER+5,A
                MOV     NVM_BUFFER+6,A
                CALL    WRITE_NVM       Write empty data to NVM
                JNC     .DONE           No write error
                MOV     ERROR,#16       Exit with write error
.DONE           RET

.NVM_NOT_EQUAL  DJNZ    R3,.NVM_SCAN_LOOP   Read following value
                RET                     Should never come here!

;---------------------------------------------------------------------------
;   ROUTINE    :  INS_VALUE
;   FUNCTION   :  Insert the value in PARAMETERS+1..+5 to memory and NVM
;---------------------------------------------------------------------------

INS_VALUE       MOV     DPTR,#VALUE_TABLE  Point to the value table
                MOV     R2,#50          Scan 50 values to see if duplicate
.MAIN_LOOP      MOV     R0,#PARAMETERS+1 Point to value to look for
                MOV     R3,DPL          Save current DPTR address
                MOV     R1,#5           Compare 5 bytes
.SCAN_LOOP      MOVX    A,@DPTR         Are they the same?
                XRL     A,@R0
                JNZ     .NOT_EQUAL      No!
                INC     R0              Test next bytes too
                INC     DPTR
                DJNZ    R1,.SCAN_LOOP
                MOV     DPL,R3          Found it!
                MOV     ERROR,#15       Exit with duplicate value error
                RET
.NOT_EQUAL      MOV     A,R3            Try next value
                ADD     A,#5
                MOV     R3,A
                MOV     DPL,A
                DJNZ    R2,.MAIN_LOOP   Try all 50 values

                MOV     DPTR,#VALUE_TABLE+4 Try to find an empty spot
                MOV     R2,#50               in the table
.FIND_SPOT      MOVX    A,@DPTR         Is this one empty?
                INC     A
                JZ      .EMPTY          Yes!
                MOV     A,DPL           Add 5 to pointer
                ADD     A,#5
                MOV     DPL,A
                DJNZ    R2,.FIND_SPOT   Try again!
                MOV     ERROR,#17       Exit with memory full error
                RET
.EMPTY          MOV     A,DPL           Subtract 4 to point to begin of
                CLR     C                value
                SUBB    A,#4
                MOV     DPL,A
                MOV     R2,#5           Save parameter value to memory
                MOV     R0,#PARAMETERS+1
.COPY_2_RAM     MOV     A,@R0           Get a byte
                MOVX    @DPTR,A          and store it
                INC     R0
                INC     DPTR
                DJNZ    R2,.COPY_2_RAM  Do all 5 bytes

.FIND_SPACE     MOV     NVM_BUFFER,#5   Read 5 bytes at a time and
                MOV     NVM_BUFFER+1,#NVM_ARRAY  start at the top of array
                MOV     R3,#50          Scan a total of 50 values
.FIND_LOOP      CALL    READ_NVM        Read a value from NVM memory
                MOV     A,NVM_BUFFER+6  Test exponent (flag) byte
                INC     A
                JZ      .NVM_EMPTY      Found a place!
                DJNZ    R3,.FIND_LOOP   Scan all 50 places
                MOV     ERROR,#17       No place found! Souldn't come here!
                RET
.NVM_EMPTY      MOV     A,NVM_BUFFER+1  Point to beginning of value
                CLR     C
                SUBB    A,#5
                MOV     NVM_BUFFER+1,A
                MOV     R0,#PARAMETERS+1 Copy the value from parameter buffer
                MOV     R1,#NVM_BUFFER+2  to the NVM buffer
                MOV     R2,#5
.COPY           MOV     A,@R0           Get byte
                MOV     @R1,A            and store it
                INC     R0
                INC     R1
                DJNZ    R2,.COPY        Do all 5 bytes
                CALL    WRITE_NVM       Write all bytes to NVM memory!
                JNC     .DONE           No write error!
                MOV     ERROR,#16       Exit with write error
.DONE           RET

;---------------------------------------------------------------------------
;   ROUTINE    :  SORT_VALUES
;   FUNCTION   :  Sort the value table
;---------------------------------------------------------------------------

SORT_VALUES     CLR     F1              Sort scaling values first
.SORT_SCALE     CALL    .SORT           Do one sort round
                JB      F0,.SORT_SCALE  Something's changed! Do it again
                SETB    F1              Sort offset values next
.SORT_OFFSET    CALL    .SORT           Do one sort round
                JB      F0,.SORT_OFFSET Something's changed! Do it again
                RET

.SORT           CLR     F0              Indicate nothing's changed
                MOV     R4,#50          Shuffle 50 values
                MOV     DPTR,#VALUE_TABLE  starting at the top of the table
                MOV     R0,#FP_ACCU_A   Get 1st value and remember it's
                CALL    .GET_VAL         start address
                MOV     R5,A
                JNC     .DONE           Not one value to sort!

.SORT_LOOP      MOV     R0,#FP_ACCU_B   Get 2nd value
                CALL    .GET_VAL
                MOV     R6,A
                JNC     .DONE           No more values to sort!
                CLR     C
                MOV     A,FP_EXP_A      Compare the 2 exponents
                SUBB    A,FP_EXP_B
                JZ      .EXP_EQUAL      Exponents are equal! Test mantissa
                JC      .ORDER_OK       Order of numbers is OK!
.SWAP           PUSH    DPL             Save current position in table
                MOV     DPL,R6          Copy Accu A to 2nd position
                CLR     C
                MOV     A,FP_EXP_A      Restore real exponent value
                SUBB    A,#10
                MOV     FP_EXP_A,A
                MOV     C,F1            Set offset/scale flag propperly
                MOV     R0,#FP_ACCU_A
                CALL    FP_ACCU_2_MEM   Copy value to memory
                MOV     DPL,R5          Copy Accu B to 1st position
                CLR     C
                MOV     A,FP_EXP_B      Restore real exponent value
                SUBB    A,#10
                MOV     FP_EXP_B,A
                MOV     C,F1            Set offset/scale flag propperly
                MOV     R0,#FP_ACCU_B
                CALL    FP_ACCU_2_MEM   Copy value to memory
                POP     DPL             Restore previous value of DPL
                MOV     A,FP_EXP_A      Re-install the exponent offset
                MOV     C,ACC.6          and sign extend to bit 7
                MOV     ACC.7,C
                ADD     A,#10
                MOV     FP_EXP_A,A
                MOV     A,R6            Data comes from new location now
                MOV     R5,A
                SETB    F0              Set Swap flag
                SJMP    .SORT_LOOP      Do next iteration
.DONE           RET

.ORDER_OK       MOV     R0,#FP_ACCU_B   Copy value of Accu B to A
                MOV     R1,#FP_ACCU_A
                CALL    COPY_FP_ACCU
                MOV     A,R6            Copy source of data pointer too
                MOV     R5,A
                SJMP    .SORT_LOOP      Do next iteration

.EXP_EQUAL      CLR     C               Subtract both mantissas from each
                MOV     R0,#FP_ACCU_A+2  other to see wich one is largest
                MOV     R1,#FP_ACCU_B+2
                MOV     R2,#4           Mantissa is 4 bytes
                CLR     SUB_ZERO        Clear zero flag
.MANTISSA_LOOP  MOV     A,@R0
                SUBB    A,@R1
                JZ      .ZERO           Value is 0
                SETB    SUB_ZERO        Set non zero flag
.ZERO           INC     R0
                INC     R1
                DJNZ    R2,.MANTISSA_LOOP  Do all bytes!
                JNB     SUB_ZERO,.ORDER_OK They were equal!? Impossible!!!
                JNC     .SWAP           Accu A was larger, so swap them
                SJMP    .ORDER_OK

.GET_VAL        MOV     A,R4            See if there are more values
                JZ      .NO_MORE        There are no more!
                PUSH    DPL             Save present pointer value
                DEC     R4
                PUSH    B0_R0           Save current pointer
                CALL    MEM_2_FP_ACCU   Get value from memory
                PUSH    PSW             Save offset/scaling flag
                MOV     A,@R0           Sign extend exponent value
                MOV     C,ACC.6
                MOV     ACC.7,C
                ADD     A,#10           Add offset to avoid negative exp
                MOV     @R0,A
                POP     PSW             Restore offset/scaling flag
                POP     B0_R0           Restore current pointer
                JB      F1,.DO_OFFSET   Searching for offset values!
                JNC     .VALUE_OK       It is a scaling factor!
.SKIP_VAL       POP     ACC             Clear stack
                SJMP    .GET_VAL        More values?
.NO_MORE        CLR     C               Indicate no more values
                RET
.DO_OFFSET      JNC     .SKIP_VAL       It's a scaling factor! Forget it
                SETB    ACC.7           See if empty value
                INC     A
                JZ      .SKIP_VAL       It's empty! Skip it
.VALUE_OK       POP     ACC             Point to beginning of this value
                SETB    C               Indicate value found
                RET

;---------------------------------------------------------------------------
;   ROUTINE    :  CALC_PERCENTAGE
;   FUNCTION   :  Calculate percentage deviation
;---------------------------------------------------------------------------

CALC_PERCENTAGE MOV     R0,#FP_ACCU_A   Copy display value back to Accu A
                MOV     DPTR,#SB_DISPLAY
                CALL    .MEM_2_FP_ACCU
                MOV     R0,#FP_ACCU_B   Get reference value in Accu B
                CALL    .GET_REF
                MOV     R0,#FP_ACCU_A   Subtract B from A
                MOV     R1,#FP_ACCU_B    Formula=  %=(freq-ref)/ref*100
                CALL    SUB_FP_NORM

                JB      SUB_SIGN,.POS   Result is positive!
                MOV     R0,#FP_ACCU_B   Copy display value back to Accu B
                MOV     DPTR,#SB_DISPLAY
                CALL    .MEM_2_FP_ACCU
                MOV     R0,#FP_ACCU_A   Get reference value in Accu A
                CALL    .GET_REF
                MOV     R0,#FP_ACCU_A   Subtract B from A
                MOV     R1,#FP_ACCU_B
                CALL    SUB_FP_NORM
                CLR     SUB_SIGN        Indicate negative number

.POS            MOV     C,SUB_SIGN      Save sign
                PUSH    PSW
                MOV     R0,#FP_ACCU_A   Copy result to Accu B
                MOV     R1,#FP_ACCU_B
                CALL    COPY_FP_ACCU
                MOV     R0,#FP_ACCU_A   Get reference value in Accu A
                CALL    .GET_REF
                CALL    DIVIDE_FP       Divide B by A
                INC     FP_EXP_C        Multiply with 100
                INC     FP_EXP_C

                MOV     R0,#FP_ACCU_C   Round Accu C to Accu B
                MOV     R1,#FP_ACCU_B
                MOV     R2,#2           Round at 2 digits as default
                MOV     A,FP_EXP_C      Round at 1 digit if val <1
                JNB     ACC.7,.ROUND2   Exponent is positive (val >=1)!
                DEC     R2              Round at 1 digit
.ROUND2         CALL    ROUND

                POP     PSW             Save sign in MSD+1 of result
                MOV     A,FP_ACCU_B+5
                MOV     ACC.7,C
                MOV     FP_ACCU_B+5,A

                MOV     R0,#FP_ACCU_B   Copy percentage value to memory
                MOV     DPTR,#PERCENT_VALUE
                CALL    FP_2_SBBUS
                MOV     R0,#FP_ACCU_B   Copy result back to Accu C
                MOV     R1,#FP_ACCU_C
                CALL    COPY_FP_ACCU
                RET

.GET_REF        MOV     DPTR,#.CCIR_REF Value in case of CCIR reference
                JNB     SPEED_REFERENCE,.CCIR  It is CCIR reference
                MOV     DPTR,#.DIN_REF  Value for DIN reference
.CCIR           MOV     R2,#7           Do a total of 7 bytes
.GET_REF_LOOP   CLR     A
                MOVC    A,@A+DPTR       Get byte from memory
                MOV     @R0,A            and copy it to RAM
                INC     R0
                INC     DPTR
                DJNZ    R2,.GET_REF_LOOP  Do all 7 bytes
                RET

.CCIR_REF       .HS     00.00.00.00.00.03.03    3000 Hz
.DIN_REF        .HS     00.00.00.00.15.03.03    3150 Hz

.MEM_2_FP_ACCU  MOV     R2,#7           Copy a total of 7 bytes
.MEM_2_FP_LOOP  MOVX    A,@DPTR         Get byte
                MOV     @R0,A            and copy it
                INC     R0
                INC     DPTR
                DJNZ    R2,.MEM_2_FP_LOOP Do all bytes
                DEC     R0              Sign extend EXP to eliminate
                MOV     C,ACC.6          6/7 digit flag
                MOV     ACC.7,C
                MOV     @R0,A
                RET

;---------------------------------------------------------------------------
;   ROUTINE    :  ROUND
;   FUNCTION   :  Round a result
;                 (R0) source FP accu, (R1) destination FP accu, R2=digits
;---------------------------------------------------------------------------

; Rounding is the process of adding 0.5 of the value of the LSD of interest,
; and then discarding everything to the right of that point.

ROUND           MOV     A,R1            Save pointer
                MOV     R4,A
                >PUSH   ACC,B0_R2       Save pointer and numer of digits
                MOV     R3,#6           Clear result accu
                CLR     A
.CLEAR          MOV     @R1,A
                INC     R1
                DJNZ    R3,.CLEAR       Do a total of 6 bytes!
                MOV     A,R0            Point to exponent of original byte
                MOV     R3,A
                ADD     A,#6
                MOV     R0,A
                MOV     A,@R0           Copy exponent to destination
                MOV     @R1,A

                MOV     A,R2            Get round (number-1)/2
                DEC     A
                CLR     C
                RRC     A
                CPL     A               Make 2s complement
                INC     A
                ADD     A,#4             and add 4 (4-(n-1)/2)
                ADD     A,R4            Add to destination pointer
                MOV     R1,A            This byte must be modified!
                MOV     A,R2            But what half of it?
                MOV     C,ACC.0
                MOV     A,#$05
                JNC     .LOW_NIB        Low nibble!
                SWAP    A
.LOW_NIB        MOV     @R1,A           Put $05 or $50 in this byte

                MOV     A,R3            Restore pointers to add original
                MOV     R1,A             value to round value
                MOV     A,R4
                MOV     R0,A
                CALL    ADD_FP_NORM

                POP     ACC             Get number of digits back
                MOV     R2,A
                DEC     A
                CLR     C
                RRC     A
                CPL     A               Make 2s complement
                INC     A
                ADD     A,#4             and add 4 (4-(n-1)/2)
                POP     B0_R0           Get destination pointer back
                JZ      .FOUND_BYTE     This is last (half) byte
.CLEAR_LOOP     MOV     @R0,#0          Clear this entire byte
                INC     R0
                DJNZ    ACC,.CLEAR_LOOP Clear all other irrelevant bytes!
.FOUND_BYTE     MOV     A,@R0           Clear at least the LSD of this byte
                ANL     A,#%1111.0000
                MOV     @R0,A
                MOV     A,R2            Must all be cleared?
                JNB     ACC.0,.DONE     No!
                MOV     @R0,#0          Clear whole byte
.DONE           RET
